
// Manages the components of the beam, and attaches it to its source
class PowerFlashlight : Powerup
{
    Property EnergyType : energyType;               // Light flickers when player is low on this inventory class
    Property FlickerThreshold : flickerThreshold;   // Threshold amount for light flicker
    Property Offset : ofsX, ofsY, ofsZ;             // Beam offset, relative to attack height

    // Beam hotspot properties
    Property SpotColor : spotColor;
    Property SpotRange : spotRange;                 // Note: I think range depends on beam angle, too
    Property SpotInnerAngle : spotInnerAngle;       // Max brightness inside this angle from center
    Property SpotOuterAngle : spotOuterAngle;       // No light outside this angle

    // Spill light properties
    Property SpillColor : spillColor;
    Property SpillRange : spillRange;
    Property SpillInnerAngle : spillInnerAngle;
    Property SpillOuterAngle : spillOuterAngle;

    // Reflected light properties
    Property BounceColor : bounceColor;
    Property BounceRange : bounceRange;             // Note: Range from point of reflection
    Property BounceInnerAngle : bounceInnerAngle;
    Property BounceOuterAngle : bounceOuterAngle;

    Default
    {
        Inventory.MaxAmount 1;
        Powerup.Duration 0x7FFFFFFF;

        PowerFlashlight.EnergyType "cell";
        PowerFlashlight.FlickerThreshold 10;
        PowerFlashlight.Offset 1, 0, 0; // ofsX is 1 so player doesn't illuminate themself

        PowerFlashlight.SpotColor 0xe3dfff;
        PowerFlashlight.SpotRange 256;
        PowerFlashlight.SpotInnerAngle 5;
        PowerFlashlight.SpotOuterAngle 20;

        PowerFlashlight.SpillColor 0xe3dfff;
        PowerFlashlight.SpillRange 0;
        PowerFlashlight.SpillInnerAngle 0;
        PowerFlashlight.SpillOuterAngle 0;

        PowerFlashlight.BounceColor 0x888888;
        PowerFlashlight.BounceRange 0; // 256;
        PowerFlashlight.BounceInnerAngle 0;
        PowerFlashlight.BounceOuterAngle 0;
    }

    class<Inventory> energyType;
    double flickerThreshold;
    double ofsX, ofsY, ofsZ;

    Color spotColor;
    int spotRange;
    double spotInnerAngle;
    double spotOuterAngle;

    Color spillColor;
    int spillRange;
    double spillInnerAngle;
    double spillOuterAngle;

    Color bounceColor;
    int bounceRange;
    double bounceInnerAngle;
    double bounceOuterAngle;

    FlashlightBeam spot;
    FlashlightBeam spill;
    FlashlightBounce bounce;

    override void AttachToOwner(Actor other)
    {
        Super.AttachToOwner(other);

        spot = FlashlightBeam(Spawn("FlashlightBeam", other.pos));
        spot.master = other;
        spot.offset = (ofsX, ofsY, ofsZ);
        spot.args[0] = spotColor.r;
        spot.args[1] = spotColor.g;
        spot.args[2] = spotColor.b;
        spot.args[3] = spotRange;
        spot.spotInnerAngle = spotInnerAngle;
        spot.spotOuterAngle = spotOuterAngle;

        if (spillRange > 0)
        {
            spill = FlashlightBeam(Spawn("FlashlightBeam", other.pos));
            spill.master = other;
            spill.offset = (ofsX, ofsY, ofsZ);
            spill.args[0] = spillColor.r;
            spill.args[1] = spillColor.g;
            spill.args[2] = spillColor.b;
            spill.args[3] = spillRange;
            spill.spotInnerAngle = spillInnerAngle;
            spill.spotOuterAngle = spillOuterAngle;
        }

        if (bounceRange > 0)
        {
            bounce = FlashlightBounce(Spawn("FlashlightBounce", other.pos));
            bounce.master = other;
            bounce.args[0] = bounceColor.r;
            bounce.args[1] = bounceColor.g;
            bounce.args[2] = bounceColor.b;
            bounce.args[3] = bounceRange;
            bounce.spotInnerAngle = bounceInnerAngle;
            bounce.spotOuterAngle = bounceOuterAngle;
        }
    }

    override void DoEffect()
    {
        Super.DoEffect();

        int energy = energyType ? owner.CountInv(energyType) : effectTics;

        int flicker = Random(0, flickerThreshold);
        if (flicker < 0) flicker *= -ticRate;

            spot.args[3] = spotRange;
            if (spill) spill.args[3] = spillRange;
            if (bounce) bounce.args[3] = bounceRange;
        
 
		//if (flicker > energy)
        //{
        //    spot.args[3] = 0;
        //    if (spill) spill.args[3] = 0;
        //    if (bounce) bounce.args[3] = 0;
        //}
        //else
        //{
        //    spot.args[3] = spotRange;
        //    if (spill) spill.args[3] = spillRange;
        //    if (bounce) bounce.args[3] = bounceRange;
        //}
    }

    override void DetachFromOwner()
    {
        spot.Destroy();
        if (spill) spill.Destroy();
        if (bounce) bounce.Destroy();
        Super.DetachFromOwner();
    }
}


// Spotlight that follows its master around
class FlashlightBeam : SpotLight
{
    Vector3 offset;

    override void Tick()
    {
        Super.Tick();

        Vector3 beamPos = master.pos;
        beamPos.xy += AngleToVector(master.angle, offset.x);
        beamPos.xy += AngleToVector(master.angle - 90, offset.y);

        let plyr = PlayerPawn(master);
        if (plyr && plyr.player)
        {
            beamPos.z += (plyr.height / 2 + plyr.attackZOffset) * plyr.player.crouchFactor;
        }
        else
        {
            beamPos.z += master.height / 2;
        }

        SetOrigin(beamPos, true);

        vel = master.vel;
        angle = master.angle;
        pitch = master.pitch;
    }
}


// Spotlight that bounces off whatever its master faces
class FlashlightBounce : SpotLight
{
    Vector3 offset;

    override void Tick()
    {
        Super.Tick();

        // Find intersection point
        double ofsZ;
        let plyr = PlayerPawn(master);
        if (plyr && plyr.player)
        {
            ofsZ += (plyr.height / 2 + plyr.attackZOffset) * plyr.player.crouchFactor;
        }
        else
        {
            ofsZ += master.height / 2;
        }

        FLineTraceData data;
        master.LineTrace(
            master.angle,
            8192,
            master.pitch,
            TRF_ThruActors | TRF_NoSky,
            ofsz, offset.x, offset.y,
            data);

        // Bounce light comes from intersection
        Vector3 bouncePos = data.hitLocation;
        bouncePos.xy += AngleToVector(master.angle + 180);
        SetOrigin(bouncePos, true);

        // Find reflection vector
        Vector3 normal;
        if (data.hitType == Trace_HitWall)
        {
            if (data.lineSide = Line.front) normal.xy = RotateVector(data.hitLine.delta.Unit(), -90);
            else normal.xy = RotateVector(data.hitLine.delta.Unit(), 90);
        }
        else if (data.hitType == Trace_HitFloor)
        {
            normal = data.hitSector.floorPlane.Normal;
        }
        else if (data.hitType == Trace_HitFloor)
        {
            normal = data.hitSector.ceilingPlane.Normal;
        }
        else
        {
            Deactivate(null);
            return;
        }

        Activate(null);

        Vector3 reflection = data.hitDir - 2 * (data.hitDir dot normal) * normal;

        angle = VectorAngle(reflection.x, reflection.y);
        pitch = -ASin(reflection.z / reflection.Length());
    }
}


// Drains player resource depending on amount in inventory
// 1 ResourceDrain = 1 unit per minute
class ResourceDrain : Inventory
{
    Property ResourceType : resourceType;

    Default
    {
        Inventory.MaxAmount 6000;
        ResourceDrain.ResourceType "Cell";
    }

    class<Inventory> resourceType;
    double spent;

    override void DoEffect()
    {
        spent += double(amount) / ticRate / 60;

        if (spent >= 1)
        {
            owner.TakeInventory(resourceType, int(spent));
            spent -= Floor(Spent);
        }
    }
}